/*********************************************************/
/* TAUCS                                                 */
/* Author: Sivan Toledo                                  */
/*********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "taucs.h"

#define max(x,y) ( ((x) > (y)) ? (x) : (y) )

/*********************************************************/
/*                                                       */
/*********************************************************/
#if 0
int
taucs_symccs_write_ijv(symccs_matrix* m, 
		       char* ijvfilename,
		       int complete)
{
  int i,ip,j,n;
  double Aij;
  FILE* f;

  f = fopen(ijvfilename , "w");

  if (f == NULL) {
    taucs_printf("symccs_write_ijv: could not open ijv file %s\n",ijvfilename);
    return -1;
  }

  n = m->n;
  
  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr[j+1]); ip++) {
      /*i   = (m->rowind)[ip] - (m->indshift);*/
      i   = (m->rowind)[ip];
      Aij = (m->values)[ip];

      fprintf(f,"%d %d %0.17e\n",i+1,j+1,Aij);
      if (i != j && complete) 
	fprintf(f,"%d %d %0.17e\n",j+1,i+1,Aij);
    }
  }

  fclose(f);

  return 0;
} 

symccs_matrix* 
taucs_symccs_read_ijv (char* ijvfilename)
{
  FILE* f;
  symccs_matrix*  m;
  int*    clen; 
  int*    is; 
  int*    js;
  double* vs;
  int ncols, nrows, nnz;
  int i,j,k,n;
  double di,dj,dv;

  f = fopen (ijvfilename , "r");

  if (f == NULL) {
    taucs_printf("symccs_read_ijv: could not open ijv file %s\n",ijvfilename);
    return NULL;
  }

  n = 10000;
  is = (int*)    malloc(n*sizeof(int));
  js = (int*)    malloc(n*sizeof(int));
  vs = (double*) malloc(n*sizeof(double));
  if (!is || !js || !vs) {
    taucs_printf("symccs_read_ijv: out of memory (1)\n");
    free(is); free(js); free(vs); 
    return NULL; 
  }


  nnz = 0;
  nrows = ncols = 0;
  while (!feof(f)) {
    if (nnz == n) {
      n = (int) ( 1.25 * (double) n);
      taucs_printf("symccs_read_ijv: allocating %d ijv's\n",n);
      is = (int*)    realloc(is,n*sizeof(int));
      js = (int*)    realloc(js,n*sizeof(int));
      vs = (double*) realloc(vs,n*sizeof(double));
      if (!is || !js || !vs) { 
	taucs_printf("symccs_read_ijv: out of memory (2)\n");
	free(is); free(js); free(vs); 
	return NULL; 
      }
    }
    if (fscanf(f, "%lg %lg %lg", &di, &dj, &dv) != 3) break;
    if (dv == 0.0) continue; /* in case the file contains explicit zeros */
    is[nnz] = di; js[nnz] = dj; vs[nnz] = dv;
    if (is[nnz] < js[nnz]) continue; /* we read the lower part */
    nrows = max(is[nnz],nrows);
    ncols = max(js[nnz],ncols);
    nnz++;
   }

  fclose ( f );

  m = (symccs_matrix*) malloc(sizeof(symccs_matrix));
  if (!m) { 
    	taucs_printf("symccs_read_ijv: out of memory (3)\n");
	free(is); free(js); free(vs); 
	return NULL; 
  }
  m->n      = nrows;
  m->uplo   = 'l';
  /*m->indshift = 0;*/
  clen      = (int*)    malloc((ncols+1) * sizeof(int));
  m->colptr = (int*)    malloc((ncols+1) * sizeof(int));
  m->rowind = (int*)    malloc(nnz       * sizeof(int));
  m->values = (double*) malloc(nnz       * sizeof(double));
  if (!clen || !(m->colptr) || !(m->rowind) || !(m->rowind)) {
    	taucs_printf("symccs_read_ijv: out of memory (4): ncols=%d nnz=%d\n",ncols,nnz);
	free(clen); free(m->colptr); free(m->rowind); free(m->values);
	free (m); free(is); free(js); free(vs); 
	return NULL; 
  }

  for (j=0; j<ncols; j++) clen[j] = 0; 
  for (k=0; k<nnz; k++) {
    i = is[k] - 1; /* make it 1-based */
    j = js[k] - 1; /* make it 1-based */
    ( clen[j] )++;
  }
  /* just check */
  k = 0;
  for (j=0; j<ncols; j++) 
    k += clen[j];
  assert(k == nnz);

  /* now compute column pointers */

  k = 0;
  for (j=0; j<ncols; j++) {
    int tmp;
    tmp =  clen[j];
    clen[j] = (m->colptr[j]) = k;
    k += tmp;
  }
  clen[ncols] = (m->colptr[ncols]) = k;
  assert(clen[ncols] == nnz);

  /* now read matrix into data structure */

  for (k=0; k<nnz; k++) {
    i = is[k] - 1; /* make it 1-based */
    j = js[k] - 1; /* make it 1-based */
    assert(i < nrows);
    assert(j < ncols);
    (m->values)[ clen[j] ] = vs[k];
    /*(m->rowind)[ clen[j] ] = i + (m->indshift);*/
    (m->rowind)[ clen[j] ] = i;
    clen[j] ++;
  }

  free(clen);
  free(vs);
  free(js);
  free(is);

  taucs_printf("symccs_read_ijv: read %s, n=%d\n",ijvfilename,m->n);

  return m;
} 
#endif
/*********************************************************/
/* CCS                                                   */
/*********************************************************/

int
taucs_ccs_write_ijv(taucs_ccs_matrix* m, 
		    char* ijvfilename)
{
  int i,ip,j,n;
  double Aij;
  FILE* f;

  f = fopen(ijvfilename , "w");

  if (f == NULL) {
    taucs_printf("taucs_ccs_write_ijv: could not open ijv file %s\n",ijvfilename);
    return -1;
  }

  n = m->n;
  
  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr[j+1]); ip++) {
      /*i   = (m->rowind)[ip] - (m->indshift);*/
      i   = (m->rowind)[ip];
      Aij = (m->values)[ip];

      fprintf(f,"%d %d %0.17e\n",i+1,j+1,Aij);
      if (i != j && ((m->flags) & TAUCS_SYMMETRIC))
	fprintf(f,"%d %d %0.17e\n",j+1,i+1,Aij);
    }
  }

  fclose(f);

  return 0;
} 

taucs_ccs_matrix* 
taucs_ccs_read_ijv(char* ijvfilename,int flags)
{
  FILE* f;
  taucs_ccs_matrix*  m;
  int*    clen; 
  int*    is; 
  int*    js;
  double* vs;
  int ncols, nrows, nnz;
  int i,j,k,n;
  double di,dj,dv;

  f = fopen (ijvfilename , "r");

  if (f == NULL) {
    taucs_printf("taucs_ccs_read_ijv: could not open ijv file %s\n",ijvfilename);
    return NULL;
  }

  n = 10000;
  is = (int*)    malloc(n*sizeof(int));
  js = (int*)    malloc(n*sizeof(int));
  vs = (double*) malloc(n*sizeof(double));
  if (!is || !js || !vs) {
    taucs_printf("symccs_read_ijv: out of memory\n");
    free(is); free(js); free(vs); 
    return NULL; 
  }

  nnz = 0;
  nrows = ncols = 0;
  while (!feof(f)) {
    if (nnz == n) {
      n = (int) ( 1.25 * (double) n);
      taucs_printf("taucs_ccs_read_ijv: allocating %d ijv's\n",n);
      is = (int*)    realloc(is,n*sizeof(int));
      js = (int*)    realloc(js,n*sizeof(int));
      vs = (double*) realloc(vs,n*sizeof(double));
      if (!is || !js || !vs) { 
	taucs_printf("taucs_ccs_read_ijv: out of memory\n");
	free(is); free(js); free(vs); 
	return NULL; 
      }
    }
    if (fscanf(f, "%lg %lg %lg", &di, &dj, &dv) != 3) break;
    is[nnz] = di; js[nnz] = dj; vs[nnz] = dv;
    /* we read the lower part */
    if ((flags & TAUCS_SYMMETRIC) && is[nnz] < js[nnz]) continue; 
    nrows = max(is[nnz],nrows);
    ncols = max(js[nnz],ncols);
    nnz++;
   }

  fclose ( f );

  m = (taucs_ccs_matrix*) malloc(sizeof(taucs_ccs_matrix));
  if (!m) { 
    taucs_printf("taucs_ccs_read_ijv: out of memory\n");
    free(is); free(js); free(vs); 
    return NULL; 
  }
  m->n      = nrows;
  m->m      = ncols;
  if (flags & TAUCS_SYMMETRIC) 
    m->flags  = TAUCS_SYMMETRIC | TAUCS_LOWER;
  else
    m->flags  = 0;
  clen      = (int*)    malloc((ncols+1) * sizeof(int));
  m->colptr = (int*)    malloc((ncols+1) * sizeof(int));
  m->rowind = (int*)    malloc(nnz       * sizeof(int));
  m->values = (double*) malloc(nnz       * sizeof(double));
  if (!clen || !(m->colptr) || !(m->rowind) || !(m->rowind)) {
    taucs_printf("taucs_ccs_read_ijv: out of memory: ncols=%d nnz=%d\n",ncols,nnz);
    free(clen); free(m->colptr); free(m->rowind); free(m->values);
    free (m); free(is); free(js); free(vs); 
    return NULL; 
  }

  for (j=0; j<ncols; j++) clen[j] = 0;
  for (k=0; k<nnz; k++) {
    i = is[k] - 1; /* make it 1-based */
    j = js[k] - 1; /* make it 1-based */
    ( clen[j] )++;
  }
  /* just check */
  k = 0;
  for (j=0; j<ncols; j++) 
    k += clen[j];
  assert(k == nnz);

  /* now compute column pointers */
  
  k = 0;
  for (j=0; j<ncols; j++) {
    int tmp;
    tmp =  clen[j];
    clen[j] = (m->colptr[j]) = k;
    k += tmp;
  }
  clen[ncols] = (m->colptr[ncols]) = k;
  assert(clen[ncols] == nnz);
  
  /* now read matrix into data structure */

  for (k=0; k<nnz; k++) {
    i = is[k] - 1; /* make it 1-based */
    j = js[k] - 1; /* make it 1-based */
    assert(i < nrows);
    assert(j < ncols);
    (m->values)[ clen[j] ] = vs[k];
    (m->rowind)[ clen[j] ] = i;
    clen[j] ++;
  }
  
  free(clen);
  free(vs);
  free(js);
  free(is);
  
  taucs_printf("taucs_ccs_read_ijv: read %s, n=%d\n",ijvfilename,m->n);

  return m;
} 

taucs_ccs_matrix* 
taucs_ccs_read_mtx(char* filename,int flags)
{
  FILE* f;
  taucs_ccs_matrix*  m;
  int*    clen; 
  int*    is; 
  int*    js;
  double* vs;
  int ncols, nrows, nnz;
  int i,j,k,n;
  double di,dj,dv;

  f = fopen (filename , "r");

  if (f == NULL) {
    taucs_printf("taucs_ccs_read_mtx: could not open mtx file %s\n",filename);
    return NULL;
  }

  if (fscanf(f, "%d %d %d", &nrows, &ncols, &nnz) != 3) {
    taucs_printf("taucs_ccs_read_mtx: wrong header\n");
    return NULL;
  }

  n = 10000;
  is = (int*)    malloc(n*sizeof(int));
  js = (int*)    malloc(n*sizeof(int));
  vs = (double*) malloc(n*sizeof(double));
  if (!is || !js || !vs) {
    taucs_printf("taucs_ccs_read_mtx: out of memory\n");
    free(is); free(js); free(vs); 
    return NULL; 
  }

  nnz = 0;
  nrows = ncols = 0;
  while (!feof(f)) {
    if (nnz == n) {
      n = (int) ( 1.25 * (double) n);
      taucs_printf("taucs_ccs_read_mtx: allocating %d ijv's\n",n);
      is = (int*)    realloc(is,n*sizeof(int));
      js = (int*)    realloc(js,n*sizeof(int));
      vs = (double*) realloc(vs,n*sizeof(double));
      if (!is || !js || !vs) { 
	taucs_printf("taucs_ccs_read_mtx: out of memory\n");
	free(is); free(js); free(vs); 
	return NULL; 
      }
    }
    if (fscanf(f, "%lg %lg %lg", &di, &dj, &dv) != 3) break;
    is[nnz] = di; js[nnz] = dj; vs[nnz] = dv;
    /* upper or lower might be stored, we use lower */
    if ((flags & TAUCS_SYMMETRIC) && is[nnz] < js[nnz]) {
      int t = is[nnz];
      is[nnz] = js[nnz];
      js[nnz] = t;
    }
    if (flags & TAUCS_PATTERN) {
      if (is[nnz] == js[nnz]) vs[nnz] = (double) (nrows+1);
      else                    vs[nnz] = -1.0;
    }
    nrows = max(is[nnz],nrows);
    ncols = max(js[nnz],ncols);
    nnz++;
   }

  fclose ( f );

  m = (taucs_ccs_matrix*) malloc(sizeof(taucs_ccs_matrix));
  if (!m) { 
    taucs_printf("taucs_ccs_read_mtx: out of memory\n");
    free(is); free(js); free(vs); 
    return NULL; 
  }
  m->n      = nrows;
  m->m      = ncols;
  if (flags & TAUCS_SYMMETRIC) 
    m->flags  = TAUCS_SYMMETRIC | TAUCS_LOWER;
  else
    m->flags  = 0;
  clen      = (int*)    malloc((ncols+1) * sizeof(int));
  m->colptr = (int*)    malloc((ncols+1) * sizeof(int));
  m->rowind = (int*)    malloc(nnz       * sizeof(int));
  m->values = (double*) malloc(nnz       * sizeof(double));
  if (!clen || !(m->colptr) || !(m->rowind) || !(m->rowind)) {
    taucs_printf("taucs_ccs_read_mtx: out of memory: ncols=%d nnz=%d\n",ncols,nnz);
    free(clen); free(m->colptr); free(m->rowind); free(m->values);
    free (m); free(is); free(js); free(vs); 
    return NULL; 
  }

  for (j=0; j<ncols; j++) clen[j] = 0;
  for (k=0; k<nnz; k++) {
    i = is[k] - 1; /* make it 1-based */
    j = js[k] - 1; /* make it 1-based */
    ( clen[j] )++;
  }
  /* just check */
  k = 0;
  for (j=0; j<ncols; j++) 
    k += clen[j];
  assert(k == nnz);

  /* now compute column pointers */
  
  k = 0;
  for (j=0; j<ncols; j++) {
    int tmp;
    tmp =  clen[j];
    clen[j] = (m->colptr[j]) = k;
    k += tmp;
  }
  clen[ncols] = (m->colptr[ncols]) = k;
  assert(clen[ncols] == nnz);
  
  /* now read matrix into data structure */

  for (k=0; k<nnz; k++) {
    i = is[k] - 1; /* make it 1-based */
    j = js[k] - 1; /* make it 1-based */
    assert(i < nrows);
    assert(j < ncols);
    (m->values)[ clen[j] ] = vs[k];
    (m->rowind)[ clen[j] ] = i;
    clen[j] ++;
  }
  
  free(clen);
  free(vs);
  free(js);
  free(is);
  
  taucs_printf("taucs_ccs_read_mtx: read %s, n=%d\n",filename,m->n);

  return m;
} 

taucs_ccs_matrix* 
taucs_ccs_read_ccs(char* filename,int flags)
{
  FILE* f;
  taucs_ccs_matrix*  m;
  int*    clen; 
  int*    is; 
  int*    js;
  double* vs;
  int ncols, nrows, nnz;
  int i,ip,j,k,n;
  double di,dj,dv;

  int N,*pointers;

  f = fopen (filename ,"r");

  if (f == NULL) {
    taucs_printf("taucs_ccs_read_ccs: could not open ccs file %s\n",filename);
    return NULL;
  }

  fscanf(f,"%d",&N);

  pointers = (int*) malloc((N+1)*sizeof(int));
  for(i=0; i<N+1; ++i) {
    fscanf(f,"%d",&pointers[i]);
  }

  m = taucs_ccs_create(N, N, pointers[N]);
  for (i=0; i<=N; i++) (m->colptr)[i] = pointers[i];

  for(i=0; i<pointers[N]; ++i)
    fscanf(f,"%d",(m->rowind)+i);
  
  for(i=0; i<pointers[N]; ++i)
    fscanf(f,"%le",(m->values)+i);
  
  if (flags & TAUCS_SYMMETRIC) {
    m->flags  = TAUCS_SYMMETRIC | TAUCS_LOWER;
    for (j=0; j<N; j++) {
      for (ip=(m->colptr)[j]; ip<(m->colptr)[j+1]; ip++) {
	i = (m->rowind)[ip];
	assert(i >= j);
      }
    }


  } else
    m->flags  = 0;


  free(pointers);
  
  taucs_printf("taucs_ccs_read_ccs: read %s, n=%d\n",filename,m->n);

  return m;
} 

/*********************************************************/
/*                                                       */
/*********************************************************/

