/*********************************************************/
/* TAUCS                                                 */
/* Author: Sivan Toledo                                  */
/*********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "taucs.h"

#define RNDM ((double)random()/(double)RAND_MAX);

#define max(x,y) ( ((x) > (y)) ? (x) : (y) )

/*********************************************************/
/* Interface to AMD                                      */
/*********************************************************/
#if 0
static void
symccs_amd(symccs_matrix* m, 
	   int** perm, int** invperm,
	   char* which)
{
#ifdef NOAMD
  taucs_printf("symccs_amd: AMD routines not linked.\n");
  *perm    = NULL;
  *invperm = NULL;
  return;
#else
  int  n, iwlen, pfree, ncmpa, iovflo;
  int* iw;
  int* pe;
  int* degree;
  int* nv;
  int* next;
  int* last;
  int* head;
  int* elen;
  int* w;
  int* len;

  int  nnz,i,j,ip;
  
  taucs_printf("symccs_amd: starting (%s)\n",which);

  *perm    = NULL;
  *invperm = NULL;

  n   = m->n;
  nnz = (m->colptr)[n];
  
  pe     = (int*) malloc(n * sizeof(int));
  degree = (int*) malloc(n * sizeof(int));
  nv     = (int*) malloc(n * sizeof(int));
  next   = (int*) malloc(n * sizeof(int));
  last   = (int*) malloc(n * sizeof(int));
  head   = (int*) malloc(n * sizeof(int));
  elen   = (int*) malloc(n * sizeof(int));
  w     = (int*) malloc(n * sizeof(int));
  len    = (int*) malloc(n * sizeof(int));

  /* AMD docs recommend iwlen >= 1.2 nnz, but this leads to compressions */
  iwlen = n + (int) (2.0 * 2.0*(nnz - n));

  taucs_printf("symccs_amd: allocating %d ints for iw\n",iwlen);

  iw = (int*) malloc(iwlen * sizeof(int));
  
  assert(iw && pe && degree && nv && next && last && head &&
	 elen && w && len); 


  iovflo = 2,147,483,648; /* for 32-bit only! */

  for (i=0; i<n; i++) len[i] = 0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      i = (m->rowind)[ip];
      /*i = (m->rowind)[ip] - (m->indshift);*/
      if (i != j) {
	len[i] ++;
	len[j] ++;
      }
    }
  }

  pe[0] = 1;
  for (i=1; i<n; i++) pe[i] = pe[i-1] + len[i-1];
  
  pfree = pe[n-1] + len[n-1];

  /* use degree as a temporary */

  for (i=0; i<n; i++) degree[i] = pe[i] - 1;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	iw[ degree[i] ] = j+1;
	iw[ degree[j] ] = i+1;
	degree[i] ++;
	degree[j] ++;
      }
    }
  }

  taucs_printf("symccs_amd: calling amd matrix is %dx%d, nnz=%d\n",
	     n,n,nnz);

  if (!strcmp(which,"mmd")) 
    amdexa_(&n, pe, iw, len, &iwlen, &pfree, nv, next,
	    last, head, elen, degree, &ncmpa, w, &iovflo);
  else if (!strcmp(which,"md")) 
    amdtru_(&n, pe, iw, len, &iwlen, &pfree, nv, next,
	  last, head, elen, degree, &ncmpa, w, &iovflo);
  else if (!strcmp(which,"amd")) 
    amdbar_(&n, pe, iw, len, &iwlen, &pfree, nv, next,
	    last, head, elen, degree, &ncmpa, w, &iovflo);
  else {
    taucs_printf("symccs_amd: WARNING - invalid ordering requested (%s)\n",which);
    return;
  }

  taucs_printf("symccs_amd: amd returned. optimal iwlen=%d (in this run was %d), %d compressions\n",
	     pfree,iwlen,ncmpa);
  /*
  {
    FILE* f;
    f=fopen("p.ijv","w");
    for (i=0; i<n; i++) fprintf(f,"%d\n",last[i]);
    fclose(f);
  }
  */

  free(pe    );
  free(degree);
  free(nv    );
  free(next  );
  /* free(last  ); */
  free(head  );
  free(elen  );
  free(w     );
  /* free(len   ); */
  free(iw    );
  
  for (i=0; i<n; i++) last[i] --;
  for (i=0; i<n; i++) len[ last[i] ] = i;

  *perm    = last;
  *invperm = len;
#endif
}

/*********************************************************/
/* Interface to MMD                                      */
/*********************************************************/

static int compare_ints(void* vx, void* vy)
{
  int* ix = (int*)vx;
  int* iy = (int*)vy;
  if (*ix < *iy) return -1;
  if (*ix > *iy) return  1;
  return 0;
}

static void
symccs_genmmd(symccs_matrix* m, 
	      int** perm, int** invperm,
	      char* which)
{
#ifdef NOMMD
  taucs_printf("symccs_mmd: MMD routines not linked.\n");
  *perm    = NULL;
  *invperm = NULL;
  return;
#else
  int  n, maxint, delta, nofsub;
  int* xadj;
  int* adjncy;
  int* invp;
  int* prm;
  int* dhead;
  int* qsize;
  int* llist;
  int* marker;

  int* len;
  int* next;

  int  nnz,i,j,ip;
  
  /*taucs_printf("symccs_genmmd: starting (%s)\n",which);*/

  *perm    = NULL;
  *invperm = NULL;

  n   = m->n;
  nnz = (m->colptr)[n];
  
  /* I copied the value of delta and the size of */
  /* from SuperLU. Sivan                         */

  delta = 1; /* DELTA is a parameter to allow the choice of nodes
		whose degree <= min-degree + DELTA. */
  delta = 1; /* DELTA is a parameter to allow the choice of nodes
		whose degree <= min-degree + DELTA. */
  maxint = 2,147,483,648; /* 2**31-1, for 32-bit only! */
  maxint = 32000;

  xadj   = (int*) malloc((n+1)     * sizeof(int));
  adjncy = (int*) malloc((2*nnz)   * sizeof(int));
  invp   = (int*) malloc((n+1)     * sizeof(int));
  prm    = (int*) malloc(n         * sizeof(int));
  dhead  = (int*) malloc((n+1)     * sizeof(int));
  qsize  = (int*) malloc((n+1)     * sizeof(int));
  llist  = (int*) malloc(n         * sizeof(int));
  marker = (int*) malloc(n         * sizeof(int));

  len  = dhead; /* we reuse space */
  next = qsize; /* we reuse space */

  for (i=0; i<n; i++) len[i] = 0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      len[i] ++;
      len[j] ++;
    }
  }

  xadj[0] = 1;
  for (i=1; i<=n; i++) xadj[i] = xadj[i-1] + len[i-1];
  
  /* use degree as a temporary */

  for (i=0; i<n; i++) next[i] = xadj[i] - 1;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      assert( next[i] < 2*nnz );
      assert( next[j] < 2*nnz );
      adjncy[ next[i] ] = j+1;
      adjncy[ next[j] ] = i+1;
      next[i] ++;
      next[j] ++;
    }
  }

  /*
  taucs_printf("symccs_genmmd: calling genmmd, matrix is %dx%d, nnz=%d\n",
	     n,n,nnz);
  */

  genmmd_(&n,
	  xadj, adjncy,
	  invp,prm,
	  &delta,
	  dhead,qsize,llist,marker,
	  &maxint,&nofsub);


  taucs_printf("symccs_genmmd: genmmd returned.\n");

  /*
  {
    FILE* f;
    f=fopen("p.ijv","w");
    for (i=0; i<n; i++) fprintf(f,"%d %d\n",prm[i],invp[i]);
    fclose(f);
  }
  */


  free(marker);
  free(llist );
  free(qsize );
  free(dhead );
  free(xadj  );
  free(adjncy);
  
  for (i=0; i<n; i++) prm[i] --;
  for (i=0; i<n; i++) invp[ prm[i] ] = i;

  *perm    = prm;
  *invperm = invp;
#endif
}

/*********************************************************/
/* No-fill ordering for trees                            */
/*********************************************************/

static void 
symccs_treeorder(symccs_matrix* m,
		 int** perm,
		 int** invperm)
{
  int  n,nnz,i,j,ip,k,p,nleaves;
  int* adjptr;
  int* adj;
  int* len;
  int* ptr;
  int* degree;
  int* leaves;

  n   = m->n;
  nnz = (m->colptr)[n];
  
  taucs_printf("symccs_treeorder: starting, matrix is %dx%d, # edges=%d\n",
	     n,n,nnz-n);

  *perm    = (int*) malloc(n * sizeof(int));
  *invperm = (int*) malloc(n * sizeof(int));

  /* we can reuse buffers: don't need invperm until the end */
  /* also, we can reuse perm for leaves but it's messy.     */
  len    = (int*) malloc(n * sizeof(int));
  degree = (int*) malloc(n * sizeof(int));
  leaves = (int*) malloc(n * sizeof(int));

  adjptr = (int*) malloc(n * sizeof(int));
  adj    = (int*) malloc(2*(nnz-n) * sizeof(int));

  if (!(*perm) || !(*invperm) || !adjptr || !adj || !len || !degree || ! leaves) {
    free(adj);
    free(adjptr);
    free(len);
    free(leaves);
    free(degree);
    free(*perm);
    free(*invperm);
    *perm = *invperm = NULL;
  }

  for (i=0; i<n; i++) len[i] = 0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	len[i] ++;
	len[j] ++;
      }
    }
  }

  nleaves = 0;
  for (i=0; i<n; i++) {
    degree[i] = len[i]; 
    if (degree[i] <= 1) {
      leaves[nleaves] = i;
      nleaves++;
    }
  }

  adjptr[0] = 0;
  for (i=1; i<n; i++) adjptr[i] = adjptr[i-1] + len[i-1];

  ptr =  *perm;
  for (i=0; i<n; i++) ptr[i] = adjptr[i];

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	adj[ ptr[i] ] = j;
	adj[ ptr[j] ] = i;
	ptr[i] ++;
	ptr[j] ++;
      }
    }
  }

  /*
  taucs_printf("symccs_treeorder: %d initial leaves: ",nleaves);
  for (i=0; i<nleaves; i++) 
    taucs_printf("%d ",leaves[i]);
  taucs_printf("\n");
  */

  for (i=0; i<n; i++) {
    nleaves--;
    if (nleaves <= 0) {
      /* not a tree */
      free(adj);
      free(adjptr);
      free(len);
      free(leaves);
      free(degree);
      free(*perm);
      free(*invperm);
      *perm = *invperm = NULL;
    }
    j = leaves[nleaves];

    /*taucs_printf("symccs_treeorder: next leaf is %d, degree=%d\n",j,len[j]);*/
    
    (*perm)   [ i ] = j;
    (*invperm)[ j ] = i;
    
    if (len[j] > 0) {
      if (len[j] != 1) {
	/* not a tree */
	free(adj);
	free(adjptr);
	free(len);
	free(leaves);
	free(degree);
	free(*perm);
	free(*invperm);
	*perm = *invperm = NULL;
      }
      p = adj[ adjptr[j] ]; 
      
      /*taucs_printf("symccs_treeorder: parent of %d is %d\n",j,p);*/

      for (k = 0; k < len[p]; k++)
	if (adj[ adjptr[p] + k ] == j) break;

      if ( k >= len[p] ) { /* otherwise j does not show up in p's adjacency list */
	/* not a tree */
	free(adj);
	free(adjptr);
	free(len);
	free(leaves);
	free(degree);
	free(*perm);
	free(*invperm);
	*perm = *invperm = NULL;
      }

      /* now delete j from p's adjacency list and compress */
      len[p] --;
      for (; k < len[p]; k++)
	adj[ adjptr[p] + k ] = adj[ adjptr[p] + k+1 ];

      if (len[p] == 1) {  /* degree was higher and now is 1 */
	leaves[ nleaves ] = p;
	nleaves++;
      }
    }
  }

  free(adj);
  free(adjptr);
  free(len);
  free(leaves);
  free(degree);

  /*
  taucs_printf("symccs_treeorder: ordering: ");
  for (i=0; i<n; i++) 
    taucs_printf("%d ",(*perm)[i]);
  taucs_printf("\n");
  */

  taucs_printf("symccs_treeorder: done\n");
}

/*********************************************************/
/* Interface to METIS                                    */
/*********************************************************/

static void 
symccs_metis(symccs_matrix* m, 
	     int** perm, int** invperm,
	     char* which)
{
#ifdef NOMETIS
  taucs_printf("symccs_metis: METIS routines not linked.\n");
  *perm    = NULL;
  *invperm = NULL;
  return;
#else
  int  n,nnz,i,j,ip;
  int* xadj;
  int* adj;
  int  num_flag     = 0;
  int  options_flag = 0;
  int* len;
  int* ptr;

  taucs_printf("symccs_metis: starting (%s)\n",which);

  n   = m->n;
  nnz = (m->colptr)[n];
  
  *perm    = (int*) malloc(n * sizeof(int));
  *invperm = (int*) malloc(n * sizeof(int));

  xadj = (int*) malloc((n+1) * sizeof(int));
  adj  = (int*) malloc(2*(nnz-n) * sizeof(int));

  assert(*perm && *invperm && xadj && adj);

  /* AMD docs recommend iwlen >= 1.2 nnz, but this leads to compressions */

  ptr = len = *perm;

  for (i=0; i<n; i++) len[i] = 0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	len[i] ++;
	len[j] ++;
      }
    }
  }

  xadj[0] = 0;
  for (i=1; i<=n; i++) xadj[i] = xadj[i-1] + len[i-1];
  
  for (i=0; i<n; i++) ptr[i] = xadj[i];

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	adj[ ptr[i] ] = j;
	adj[ ptr[j] ] = i;
	ptr[i] ++;
	ptr[j] ++;
      }
    }
  }

  taucs_printf("symccs_metis: calling metis matrix is %dx%d, nnz=%d\n",
	     n,n,nnz);

  METIS_NodeND(&n,
	       xadj,adj,
	       &num_flag, &options_flag,
	       *perm,*invperm);

  taucs_printf("symccs_metis: metis returned\n");

  /*
  {
    FILE* f;
    f=fopen("p.ijv","w");
    for (i=0; i<n; i++) fprintf(f,"%d\n",last[i]);
    fclose(f);
  }
  */

  free(xadj);
  free(adj);
#endif
}

/*********************************************************/
/* MAIN ORDERING ROUTINE                                 */
/*********************************************************/

void 
taucs_symccs_order(symccs_matrix* m, 
		   int** perm, int** invperm,
		   char* which)
{
  taucs_printf("symccs_order: starting (%s)\n",which);

  if (!strcmp(which,"mmd") || !strcmp(which,"amd") || !strcmp(which,"md")) 
    symccs_amd(m,perm,invperm,which);
  else if (!strcmp(which,"metis"))
    symccs_metis(m,perm,invperm,which);
  else if (!strcmp(which,"genmmd"))
    symccs_genmmd(m,perm,invperm,which);
  else if (!strcmp(which,"tree")) {
    symccs_treeorder(m,perm,invperm);
    if (*perm == NULL) /* perhaps the graph of the matrix is not a tree */
      symccs_metis(m,perm,invperm,"metis");
  }
  else if (!strcmp(which,"identity")) {
    int i;
    *perm    = (int*) malloc((m->n) * sizeof(int));
    *invperm = (int*) malloc((m->n) * sizeof(int));
    if (!(*perm) || !(*invperm)) {
      free(*perm); free(*invperm);
      taucs_printf("taucs_symccs_order: out of memory for identity permutation\n");
      return;
    }
    for (i=0; i<m->n; i++) (*perm)[i] = (*invperm)[i] = i;
    return;
  }
  else {
    taucs_printf("taucs_symccs_order: invalid ordering requested (%s)\n",which);
    *perm = *invperm = NULL;
  }
}
#endif 
/*********************************************************/
/* CCS                                                   */
/*********************************************************/


/*********************************************************/
/* Interface to AMD                                      */
/*********************************************************/

static void
taucs_ccs_amd(taucs_ccs_matrix* m, 
	      int** perm, int** invperm,
	      char* which)
{
#ifdef NOAMD
  taucs_printf("symccs_amd: AMD routines not linked.\n");
  *perm    = NULL;
  *invperm = NULL;
  return;
#else
  int  n, iwlen, pfree, ncmpa, iovflo;
  int* iw;
  int* pe;
  int* degree;
  int* nv;
  int* next;
  int* last;
  int* head;
  int* elen;
  int* w;
  int* len;

  int  nnz,i,j,ip;
  
  taucs_printf("taucs_ccs_amd: starting (%s)\n",which);

  if (!(m->flags & TAUCS_SYMMETRIC)) {
    taucs_printf("taucs_ccs_amd: AMD ordering only works on symmetric matrices.\n");
    *perm    = NULL;
    *invperm = NULL;
    return;
  }
  /* this routine may actually work on UPPER as well */
  if (!(m->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_amd: the lower part of the matrix must be represented.\n");
    *perm    = NULL;
    *invperm = NULL;
    return;
  }
    
  *perm    = NULL;
  *invperm = NULL;

  n   = m->n;
  nnz = (m->colptr)[n];
  
  pe     = (int*) malloc(n * sizeof(int));
  degree = (int*) malloc(n * sizeof(int));
  nv     = (int*) malloc(n * sizeof(int));
  next   = (int*) malloc(n * sizeof(int));
  last   = (int*) malloc(n * sizeof(int));
  head   = (int*) malloc(n * sizeof(int));
  elen   = (int*) malloc(n * sizeof(int));
  w     = (int*) malloc(n * sizeof(int));
  len    = (int*) malloc(n * sizeof(int));

  /* AMD docs recommend iwlen >= 1.2 nnz, but this leads to compressions */
  iwlen = n + (int) (2.0 * 2.0*(nnz - n));

  taucs_printf("taucs_ccs_amd: allocating %d ints for iw\n",iwlen);

  iw = (int*) malloc(iwlen * sizeof(int));
  
  assert(iw && pe && degree && nv && next && last && head &&
	 elen && w && len); 


  iovflo = 2,147,483,648; /* for 32-bit only! */

  for (i=0; i<n; i++) len[i] = 0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      i = (m->rowind)[ip];
      /*i = (m->rowind)[ip] - (m->indshift);*/
      if (i != j) {
	len[i] ++;
	len[j] ++;
      }
    }
  }

  pe[0] = 1;
  for (i=1; i<n; i++) pe[i] = pe[i-1] + len[i-1];
  
  pfree = pe[n-1] + len[n-1];

  /* use degree as a temporary */

  for (i=0; i<n; i++) degree[i] = pe[i] - 1;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	iw[ degree[i] ] = j+1;
	iw[ degree[j] ] = i+1;
	degree[i] ++;
	degree[j] ++;
      }
    }
  }

  taucs_printf("taucs_ccs_amd: calling amd matrix is %dx%d, nnz=%d\n",
	     n,n,nnz);

  if (!strcmp(which,"mmd")) 
    amdexa_(&n, pe, iw, len, &iwlen, &pfree, nv, next,
	    last, head, elen, degree, &ncmpa, w, &iovflo);
  else if (!strcmp(which,"md")) 
    amdtru_(&n, pe, iw, len, &iwlen, &pfree, nv, next,
	  last, head, elen, degree, &ncmpa, w, &iovflo);
  else if (!strcmp(which,"amd")) 
    amdbar_(&n, pe, iw, len, &iwlen, &pfree, nv, next,
	    last, head, elen, degree, &ncmpa, w, &iovflo);
  else {
    taucs_printf("taucs_ccs_amd: WARNING - invalid ordering requested (%s)\n",which);
    return;
  }

  taucs_printf("taucs_ccs_amd: amd returned. optimal iwlen=%d (in this run was %d), %d compressions\n",
	     pfree,iwlen,ncmpa);
  /*
  {
    FILE* f;
    f=fopen("p.ijv","w");
    for (i=0; i<n; i++) fprintf(f,"%d\n",last[i]);
    fclose(f);
  }
  */

  free(pe    );
  free(degree);
  free(nv    );
  free(next  );
  /* free(last  ); */
  free(head  );
  free(elen  );
  free(w     );
  /* free(len   ); */
  free(iw    );
  
  for (i=0; i<n; i++) last[i] --;
  for (i=0; i<n; i++) len[ last[i] ] = i;

  *perm    = last;
  *invperm = len;
#endif
}

/*********************************************************/
/* Interface to MMD                                      */
/*********************************************************/

static void
taucs_ccs_genmmd(taucs_ccs_matrix* m, 
		 int** perm, int** invperm,
		 char* which)
{
#ifdef NOMMD
  taucs_printf("taucs_ccs_genmmd: GENMMD routines not linked.\n");
  *perm    = NULL;
  *invperm = NULL;
  return;
#else
  int  n, maxint, delta, nofsub;
  int* xadj;
  int* adjncy;
  int* invp;
  int* prm;
  int* dhead;
  int* qsize;
  int* llist;
  int* marker;

  int* len;
  int* next;

  int  nnz,i,j,ip;
  
  /*taucs_printf("taucs_ccs_genmmd: starting (%s)\n",which);*/

  if (!(m->flags & TAUCS_SYMMETRIC)) {
    taucs_printf("taucs_ccs_genmmd: GENMMD ordering only works on symmetric matrices.\n");
    *perm    = NULL;
    *invperm = NULL;
    return;
  }
  /* this routine may actually work on UPPER as well */
  if (!(m->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_genmmd: the lower part of the matrix must be represented.\n");
    *perm    = NULL;
    *invperm = NULL;
    return;
  }

  *perm    = NULL;
  *invperm = NULL;

  n   = m->n;
  nnz = (m->colptr)[n];
  
  /* I copied the value of delta and the size of */
  /* from SuperLU. Sivan                         */

  delta = 1; /* DELTA is a parameter to allow the choice of nodes
		whose degree <= min-degree + DELTA. */
  delta = 1; /* DELTA is a parameter to allow the choice of nodes
		whose degree <= min-degree + DELTA. */
  maxint = 2,147,483,648; /* 2**31-1, for 32-bit only! */
  maxint = 32000;

  xadj   = (int*) malloc((n+1)     * sizeof(int));
  adjncy = (int*) malloc((2*nnz-n) * sizeof(int));
  invp   = (int*) malloc((n+1)     * sizeof(int));
  prm    = (int*) malloc(n         * sizeof(int));
  dhead  = (int*) malloc((n+1)     * sizeof(int));
  qsize  = (int*) malloc((n+1)     * sizeof(int));
  llist  = (int*) malloc(n         * sizeof(int));
  marker = (int*) malloc(n         * sizeof(int));

  len  = dhead; /* we reuse space */
  next = qsize; /* we reuse space */

  for (i=0; i<n; i++) len[i] = 0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	len[i] ++;
	len[j] ++;
      } else {
	/*len[i] ++;*/
      }
    }
  }

  xadj[0] = 1;
  for (i=1; i<=n; i++) xadj[i] = xadj[i-1] + len[i-1];

  /*for (i=0; i<=n; i++) printf("xadj[%d]=%d\n",i,xadj[i]);*/
  
  /* use degree as a temporary */

  for (i=0; i<n; i++) next[i] = xadj[i] - 1;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      assert( next[i] < 2*nnz-n );
      assert( next[j] < 2*nnz-n );
      if (i != j) {
	adjncy[ next[i] ] = j+1;
	adjncy[ next[j] ] = i+1;
	next[i] ++;
	next[j] ++;
      } else {
	/*
        adjncy[ next[i] ] = j+1;
	next[i] ++;
	*/
      }
    }
  }

  /*
  for (j=0; j<n; j++) {
    qsort(adjncy + (xadj[j] - 1),
	  xadj[j+1] - xadj[j],
	  sizeof(int),
	  compare_ints);
    printf("+++ %d: ",j+1);
    for (ip=xadj[j]-1; ip<xadj[j+1]-1;ip++)
      printf("%d ",adjncy[ip]);
    printf("\n");
  }
  */

  /*
  taucs_printf("taucs_ccs_genmmd: calling genmmd, matrix is %dx%d, nnz=%d\n",
	     n,n,nnz);
  */

  genmmd_(&n,
	  xadj, adjncy,
	  invp,prm,
	  &delta,
	  dhead,qsize,llist,marker,
	  &maxint,&nofsub);


  /*taucs_printf("taucs_ccs_genmmd: genmmd returned.\n");*/

  /*
  {
    FILE* f;
    f=fopen("p.ijv","w");
    for (i=0; i<n; i++) fprintf(f,"%d %d\n",prm[i],invp[i]);
    fclose(f);
  }
  */


  free(marker);
  free(llist );
  free(qsize );
  free(dhead );
  free(xadj  );
  free(adjncy);
  
  for (i=0; i<n; i++) prm[i] --;
  for (i=0; i<n; i++) invp[ prm[i] ] = i;

  *perm    = prm;
  *invperm = invp;
#endif
}

/*********************************************************/
/* No-fill ordering for trees                            */
/*********************************************************/

static void 
taucs_ccs_treeorder(taucs_ccs_matrix* m,
		    int** perm,
		    int** invperm)
{
  int  n,nnz,i,j,ip,k,p,nleaves;
  int* adjptr;
  int* adj;
  int* len;
  int* ptr;
  int* degree;
  int* leaves;

  if (!(m->flags & TAUCS_SYMMETRIC)) {
    taucs_printf("taucs_ccs_treeorder: tree ordering only works on symmetric matrices.\n");
    *perm    = NULL;
    *invperm = NULL;
    return;
  }
  /* this routine may actually work on UPPER as well */
  if (!(m->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_treeorder: the lower part of the matrix must be represented.\n");
    *perm    = NULL;
    *invperm = NULL;
    return;
  }

  n   = m->n;
  nnz = (m->colptr)[n];
  
  taucs_printf("taucs_ccs_treeorder: starting, matrix is %dx%d, # edges=%d\n",
	     n,n,nnz-n);

  *perm    = (int*) malloc(n * sizeof(int));
  *invperm = (int*) malloc(n * sizeof(int));

  /* we can reuse buffers: don't need invperm until the end */
  /* also, we can reuse perm for leaves but it's messy.     */
  len    = (int*) malloc(n * sizeof(int));
  degree = (int*) malloc(n * sizeof(int));
  leaves = (int*) malloc(n * sizeof(int));

  adjptr = (int*) malloc(n * sizeof(int));
  adj    = (int*) malloc(2*(nnz-n) * sizeof(int));

  if (!(*perm) || !(*invperm) || !adjptr || !adj || !len || !degree || ! leaves) {
    free(adj);
    free(adjptr);
    free(len);
    free(leaves);
    free(degree);
    free(*perm);
    free(*invperm);
    *perm = *invperm = NULL;
  }

  for (i=0; i<n; i++) len[i] = 0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	len[i] ++;
	len[j] ++;
      }
    }
  }

  nleaves = 0;
  for (i=0; i<n; i++) {
    degree[i] = len[i]; 
    if (degree[i] <= 1) {
      leaves[nleaves] = i;
      nleaves++;
    }
  }

  adjptr[0] = 0;
  for (i=1; i<n; i++) adjptr[i] = adjptr[i-1] + len[i-1];

  ptr =  *perm;
  for (i=0; i<n; i++) ptr[i] = adjptr[i];

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	adj[ ptr[i] ] = j;
	adj[ ptr[j] ] = i;
	ptr[i] ++;
	ptr[j] ++;
      }
    }
  }

  /*
  taucs_printf("taucs_ccs_treeorder: %d initial leaves: ",nleaves);
  for (i=0; i<nleaves; i++) 
    taucs_printf("%d ",leaves[i]);
  taucs_printf("\n");
  */

  for (i=0; i<n; i++) {
    nleaves--;
    if (nleaves <= 0) {
      /* not a tree */
      free(adj);
      free(adjptr);
      free(len);
      free(leaves);
      free(degree);
      free(*perm);
      free(*invperm);
      *perm = *invperm = NULL;
    }
    j = leaves[nleaves];

    /*taucs_printf("taucs_ccs_treeorder: next leaf is %d, degree=%d\n",j,len[j]);*/
    
    (*perm)   [ i ] = j;
    (*invperm)[ j ] = i;
    
    if (len[j] > 0) {
      if (len[j] != 1) {
	/* not a tree */
	free(adj);
	free(adjptr);
	free(len);
	free(leaves);
	free(degree);
	free(*perm);
	free(*invperm);
	*perm = *invperm = NULL;
      }
      p = adj[ adjptr[j] ]; 
      
      /*taucs_printf("taucs_ccs_treeorder: parent of %d is %d\n",j,p);*/

      for (k = 0; k < len[p]; k++)
	if (adj[ adjptr[p] + k ] == j) break;

      if ( k >= len[p] ) { /* otherwise j does not show up in p's adjacency list */
	/* not a tree */
	free(adj);
	free(adjptr);
	free(len);
	free(leaves);
	free(degree);
	free(*perm);
	free(*invperm);
	*perm = *invperm = NULL;
      }

      /* now delete j from p's adjacency list and compress */
      len[p] --;
      for (; k < len[p]; k++)
	adj[ adjptr[p] + k ] = adj[ adjptr[p] + k+1 ];

      if (len[p] == 1) {  /* degree was higher and now is 1 */
	leaves[ nleaves ] = p;
	nleaves++;
      }
    }
  }

  free(adj);
  free(adjptr);
  free(len);
  free(leaves);
  free(degree);

  /*
  taucs_printf("taucs_ccs_treeorder: ordering: ");
  for (i=0; i<n; i++) 
    taucs_printf("%d ",(*perm)[i]);
  taucs_printf("\n");
  */

  taucs_printf("taucs_ccs_treeorder: done\n");
}

/*********************************************************/
/* Interface to METIS                                    */
/*********************************************************/

static void 
taucs_ccs_metis(taucs_ccs_matrix* m, 
		int** perm, int** invperm,
		char* which)
{
#ifdef NOMETIS
  taucs_printf("taucs_ccs_metis: METIS routines not linked.\n");
  *perm    = NULL;
  *invperm = NULL;
  return;
#else
  int  n,nnz,i,j,ip;
  int* xadj;
  int* adj;
  int  num_flag     = 0;
  int  options_flag = 0;
  int* len;
  int* ptr;

  /* taucs_printf("taucs_ccs_metis: starting (%s)\n",which); */

  if (!(m->flags & TAUCS_SYMMETRIC)) {
    taucs_printf("taucs_ccs_treeorder: METIS ordering only works on symmetric matrices.\n");
    *perm    = NULL;
    *invperm = NULL;
    return;
  }
  /* this routine may actually work on UPPER as well */
  if (!(m->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_metis: the lower part of the matrix must be represented.\n");
    *perm    = NULL;
    *invperm = NULL;
    return;
  }

  n   = m->n;
  nnz = (m->colptr)[n];
  
  *perm    = (int*) malloc(n * sizeof(int));
  *invperm = (int*) malloc(n * sizeof(int));

  xadj = (int*) malloc((n+1) * sizeof(int));
  adj  = (int*) malloc(2*(nnz-n) * sizeof(int));

  assert(*perm && *invperm && xadj && adj);

  /* AMD docs recommend iwlen >= 1.2 nnz, but this leads to compressions */

  ptr = len = *perm;

  for (i=0; i<n; i++) len[i] = 0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	len[i] ++;
	len[j] ++;
      }
    }
  }

  xadj[0] = 0;
  for (i=1; i<=n; i++) xadj[i] = xadj[i-1] + len[i-1];
  
  for (i=0; i<n; i++) ptr[i] = xadj[i];

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr)[j+1]; ip++) {
      /*i = (m->rowind)[ip] - (m->indshift);*/
      i = (m->rowind)[ip];
      if (i != j) {
	adj[ ptr[i] ] = j;
	adj[ ptr[j] ] = i;
	ptr[i] ++;
	ptr[j] ++;
      }
    }
  }

  /* taucs_printf("taucs_ccs_metis: calling metis matrix is %dx%d, nnz=%d\n", */
	     /* n,n,nnz); */

  METIS_NodeND(&n,
	       xadj,adj,
	       &num_flag, &options_flag,
	       *perm,*invperm);

  /* taucs_printf("taucs_ccs_metis: metis returned\n"); */

  /*
  {
    FILE* f;
    f=fopen("p.ijv","w");
    for (i=0; i<n; i++) fprintf(f,"%d\n",last[i]);
    fclose(f);
  }
  */

  free(xadj);
  free(adj);
#endif
}

/*********************************************************/
/* MAIN ORDERING ROUTINE                                 */
/*********************************************************/

void 
taucs_ccs_order(taucs_ccs_matrix* m, 
		int** perm, int** invperm,
		char* which)
{
  if (!strcmp(which,"mmd") || !strcmp(which,"amd") || !strcmp(which,"md")) 
    taucs_ccs_amd(m,perm,invperm,which);
  else if (!strcmp(which,"metis"))
    taucs_ccs_metis(m,perm,invperm,which);
  else if (!strcmp(which,"genmmd"))
    taucs_ccs_genmmd(m,perm,invperm,which);
  else if (!strcmp(which,"tree")) {
    taucs_ccs_treeorder(m,perm,invperm);
    if (*perm == NULL) /* perhaps the graph of the matrix is not a tree */
      taucs_ccs_metis(m,perm,invperm,"metis");
  }
  else if (!strcmp(which,"identity")) {
    int i;
    *perm    = (int*) malloc((m->n) * sizeof(int));
    *invperm = (int*) malloc((m->n) * sizeof(int));
    if (!(*perm) || !(*invperm)) {
      free(*perm); free(*invperm);
      taucs_printf("taucs_ccs_order: out of memory for identity permutation\n");
      return;
    }
    for (i=0; i<m->n; i++) (*perm)[i] = (*invperm)[i] = i;
    return;
  }
  else {
    taucs_printf("taucs_ccs_order: invalid ordering requested (%s)\n",which);
    *perm = *invperm = NULL;
  }
}

/*********************************************************/
/*                                                       */
/*********************************************************/


